#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module implements the utilities for CKIP Transformers NLP drivers.
"""

__author__ = "Mu Yang <http://muyang.pro>"
__copyright__ = "2021 CKIP Lab"
__license__ = "GPL-3.0"


from abc import (
    ABCMeta,
    abstractmethod,
)

from typing import (
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

from tqdm import tqdm

import numpy as np

import torch
from torch.utils.data import (
    DataLoader,
    TensorDataset,
)

from transformers import (
    AutoModelForTokenClassification,
    BatchEncoding,
    BertTokenizerFast,
)

################################################################################################################################


class CkipTokenClassification(metaclass=ABCMeta):
    """The base class for token classification task.

    Parameters
    ----------
        model_name : ``str``
            The pretrained model name (e.g. ``'ckiplab/bert-base-chinese-ws'``).
        tokenizer_name : ``str``, *optional*, defaults to **model_name**
            The pretrained tokenizer name (e.g. ``'bert-base-chinese'``).
        device : ``int``, *optional*, defaults to -1
            Device ordinal for CPU/GPU supports.
            Setting this to -1 will leverage CPU, a positive will run the model on the associated CUDA device id.
    """

    def __init__(
        self,
        model_name: str,
        tokenizer_name: Optional[str] = None,
        *,
        device: int = -1,
    ):
        self.model = AutoModelForTokenClassification.from_pretrained(model_name)
        self.tokenizer = BertTokenizerFast.from_pretrained(tokenizer_name or model_name)

        self.device = torch.device("cpu" if device < 0 else f"cuda:{device}")  # pylint: disable=no-member
        self.model.to(self.device)

    ########################################################################################################################

    @classmethod
    @abstractmethod
    def _model_names(cls):
        return NotImplemented  # pragma: no cover

    def _get_model_name(
        self,
        model: str,
    ):
        try:
            model_name = self._model_names[model]
        except KeyError as exc:
            raise KeyError(f"Invalid model {model}") from exc

        return model_name

    ########################################################################################################################

    def __call__(
        self,
        input_text: Union[List[str], List[List[str]]],
        *,
        use_delim: bool = False,
        delim_set: Optional[str] = "，,。：:；;！!？?",
        batch_size: int = 256,
        max_length: Optional[int] = None,
        show_progress: bool = True,
        pin_memory: bool = True,
    ):
        """Call the driver.

        Parameters
        ----------
            input_text : ``List[str]`` or ``List[List[str]]``
                The input sentences. Each sentence is a string or a list of string.
            use_delim : ``bool``, *optional*, defaults to False
                Segment sentence (internally) using ``delim_set``.
            delim_set : `str`, *optional*, defaults to ``'，,。：:；;！!？?'``
                Used for sentence segmentation if ``use_delim=True``.
            batch_size : ``int``, *optional*, defaults to 256
                The size of mini-batch.
            max_length : ``int``, *optional*
                The maximum length of the sentence,
                must not longer then the maximum sequence length for this model (i.e. ``tokenizer.model_max_length``).
            show_progress : ``bool``, *optional*, defaults to True
                Show progress bar.
            pin_memory : ``bool``, *optional*, defaults to True
                Pin memory in order to accelerate the speed of data transfer to the GPU. This option is
                incompatible with multiprocessing.
        """

        model_max_length = self.tokenizer.model_max_length - 2  # Add [CLS] and [SEP]
        if max_length:
            assert max_length < model_max_length, (
                "Sequence length is longer than the maximum sequence length for this model "
                f"({max_length} > {model_max_length})."
            )
        else:
            max_length = model_max_length

        # Apply delimiter cut
        delim_index = self._find_delim(
            input_text=input_text,
            use_delim=use_delim,
            delim_set=delim_set,
        )

        # Get worded input IDs
        if show_progress:
            input_text = tqdm(input_text, desc="Tokenization")

        input_ids_worded = [
            [self.tokenizer.convert_tokens_to_ids(list(input_word)) for input_word in input_sent] for input_sent in input_text
        ]

        # Flatten input IDs
        (input_ids, index_map,) = self._flatten_input_ids(
            input_ids_worded=input_ids_worded,
            max_length=max_length,
            delim_index=delim_index,
        )

        # Pad and segment input IDs
        (input_ids, attention_mask,) = self._pad_input_ids(
            input_ids=input_ids,
        )

        # Convert input format
        encoded_input = BatchEncoding(
            data=dict(
                input_ids=input_ids,
                attention_mask=attention_mask,
            ),
            tensor_type="pt",
        )

        # Create dataset
        dataset = TensorDataset(*encoded_input.values())
        dataloader = DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=False,
            drop_last=False,
            pin_memory=pin_memory,
        )
        if show_progress:
            dataloader = tqdm(dataloader, desc="Inference")

        # Call Model
        logits = []
        with torch.no_grad():
            for batch in dataloader:
                batch = tuple(tensor.to(self.device) for tensor in batch)
                (batch_logits,) = self.model(**dict(zip(encoded_input.keys(), batch)), return_dict=False)
                batch_logits = batch_logits.cpu().numpy()[:, 1:, :]  # Remove [CLS]
                logits.append(batch_logits)

        # Call model
        logits = np.concatenate(logits, axis=0)

        return logits, index_map

    @staticmethod
    def _find_delim(
        *,
        input_text,
        use_delim,
        delim_set,
    ):
        if not use_delim:
            return set()

        delim_index = set()
        delim_set = set(delim_set)
        for sent_idx, input_sent in enumerate(input_text):
            for word_idx, input_word in enumerate(input_sent):
                if input_word in delim_set:
                    delim_index.add((sent_idx, word_idx))
        return delim_index

    @staticmethod
    def _flatten_input_ids(
        *,
        input_ids_worded,
        max_length,
        delim_index,
    ):
        input_ids = []
        index_map = []

        input_ids_sent = []
        index_map_sent = []

        for sent_idx, input_ids_worded_sent in enumerate(input_ids_worded):
            for word_idx, word_ids in enumerate(input_ids_worded_sent):
                word_length = len(word_ids)

                if word_length == 0:
                    index_map_sent.append(None)
                    continue

                # Check if sentence segmentation is needed
                if len(input_ids_sent) + word_length > max_length:
                    input_ids.append(input_ids_sent)
                    input_ids_sent = []

                # Insert tokens
                index_map_sent.append(
                    (
                        len(input_ids),  # line index
                        len(input_ids_sent),  # token index
                    )
                )
                input_ids_sent += word_ids

                if (sent_idx, word_idx) in delim_index:
                    input_ids.append(input_ids_sent)
                    input_ids_sent = []

            # End of a sentence
            if input_ids_sent:
                input_ids.append(input_ids_sent)
                input_ids_sent = []
            index_map.append(index_map_sent)
            index_map_sent = []

        return input_ids, index_map

    def _pad_input_ids(
        self,
        *,
        input_ids,
    ):
        max_length = max(map(len, input_ids))

        padded_input_ids = []
        attention_mask = []
        for input_ids_sent in input_ids:
            token_count = len(input_ids_sent)
            pad_count = max_length - token_count
            padded_input_ids.append(
                [self.tokenizer.cls_token_id]
                + input_ids_sent
                + [self.tokenizer.sep_token_id]
                + [self.tokenizer.pad_token_id] * pad_count
            )
            attention_mask.append([1] * (token_count + 2) + [0] * pad_count)  # [CLS] & input & [SEP]  # [PAD]s
        return padded_input_ids, attention_mask


################################################################################################################################


class NerToken(NamedTuple):
    """A named-entity recognition token."""

    word: str  #: ``str``, the token word.
    ner: str  #: ``str``, the NER-tag.
    idx: Tuple[int, int]  #: ``Tuple[int, int]``, the starting / ending index in the sentence.



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides the CKIP Transformers NLP drivers.
"""

__author__ = "Mu Yang <http://muyang.pro>"
__copyright__ = "2021 CKIP Lab"
__license__ = "GPL-3.0"

from .driver import (
    CkipWordSegmenter,
    CkipPosTagger,
    CkipNerChunker,
)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module implements the CKIP Transformers NLP drivers.
"""

__author__ = "Mu Yang <http://muyang.pro>"
__copyright__ = "2021 CKIP Lab"
__license__ = "GPL-3.0"

from typing import (
    List,
)

import numpy as np

from .util import (
    CkipTokenClassification,
    NerToken,
)

################################################################################################################################


class CkipWordSegmenter(CkipTokenClassification):
    """The word segmentation driver.

    Parameters
    ----------
        model : ``str`` *optional*, defaults to "bert-base".
            The pretrained model name provided by CKIP Transformers.
        model_name : ``str`` *optional*, overwrites **model**
            The custom pretrained model name (e.g. ``'ckiplab/bert-base-chinese-ws'``).
        device : ``int``, *optional*, defaults to -1
            Device ordinal for CPU/GPU supports.
            Setting this to -1 will leverage CPU, a positive will run the model on the associated CUDA device id.
    """

    _model_names = {
        "albert-tiny": "ckiplab/albert-tiny-chinese-ws",
        "albert-base": "ckiplab/albert-base-chinese-ws",
        "bert-tiny": "ckiplab/bert-tiny-chinese-ws",
        "bert-base": "ckiplab/bert-base-chinese-ws",
    }

    def __init__(
        self,
        model: str = "bert-base",
        **kwargs,
    ):
        model_name = kwargs.pop("model_name", self._get_model_name(model))
        super().__init__(model_name=model_name, **kwargs)

    def __call__(
        self,
        input_text: List[str],
        *,
        use_delim: bool = False,
        **kwargs,
    ) -> List[List[str]]:
        """Call the driver.

        Parameters
        ----------
            input_text : ``List[str]``
                The input sentences. Each sentence is a string.
            use_delim : ``bool``, *optional*, defaults to False
                Segment sentence (internally) using ``delim_set``.
            delim_set : `str`, *optional*, defaults to ``'，,。：:；;！!？?'``
                Used for sentence segmentation if ``use_delim=True``.
            batch_size : ``int``, *optional*, defaults to 256
                The size of mini-batch.
            max_length : ``int``, *optional*
                The maximum length of the sentence,
                must not longer then the maximum sequence length for this model (i.e. ``tokenizer.model_max_length``).
            show_progress : ``int``, *optional*, defaults to True
                Show progress bar.
            pin_memory : ``bool``, *optional*, defaults to True
                Pin memory in order to accelerate the speed of data transfer to the GPU. This option is
                incompatible with multiprocessing.

        Returns
        -------
            ``List[List[str]]``
                A list of list of words (``str``).
        """

        # Call model
        (
            logits,
            index_map,
        ) = super().__call__(input_text, use_delim=use_delim, **kwargs)

        # Post-process results
        output_text = []
        for sent_data in zip(input_text, index_map):
            output_sent = []
            word = ""
            for input_char, logits_index in zip(*sent_data):
                if logits_index is None:
                    if word:
                        output_sent.append(word)
                    output_sent.append(input_char)
                    word = ""
                else:
                    logits_b, logits_i = logits[logits_index]

                    if logits_b > logits_i:
                        if word:
                            output_sent.append(word)
                        word = input_char
                    else:
                        word += input_char

            if word:
                output_sent.append(word)
            output_text.append(output_sent)

        return output_text


################################################################################################################################


class CkipPosTagger(CkipTokenClassification):
    """The part-of-speech tagging driver.

    Parameters
    ----------
        model : ``str`` *optional*, defaults to "bert-base".
            The pretrained model name provided by CKIP Transformers.
        model_name : ``str`` *optional*, overwrites **model**
            The custom pretrained model name (e.g. ``'ckiplab/bert-base-chinese-pos'``).
        device : ``int``, *optional*, defaults to -1
            Device ordinal for CPU/GPU supports.
            Setting this to -1 will leverage CPU, a positive will run the model on the associated CUDA device id.
    """

    _model_names = {
        "albert-tiny": "ckiplab/albert-tiny-chinese-pos",
        "albert-base": "ckiplab/albert-base-chinese-pos",
        "bert-tiny": "ckiplab/bert-tiny-chinese-pos",
        "bert-base": "ckiplab/bert-base-chinese-pos",
    }

    def __init__(
        self,
        model: str = "bert-base",
        **kwargs,
    ):
        model_name = kwargs.pop("model_name", self._get_model_name(model))
        super().__init__(model_name=model_name, **kwargs)

    def __call__(
        self,
        input_text: List[List[str]],
        *,
        use_delim: bool = True,
        **kwargs,
    ) -> List[List[str]]:
        """Call the driver.

        Parameters
        ----------
            input_text : ``List[List[str]]``
                The input sentences. Each sentence is a list of strings (words).
            use_delim : ``bool``, *optional*, defaults to True
                Segment sentence (internally) using ``delim_set``.
            delim_set : `str`, *optional*, defaults to ``'，,。：:；;！!？?'``
                Used for sentence segmentation if ``use_delim=True``.
            batch_size : ``int``, *optional*, defaults to 256
                The size of mini-batch.
            max_length : ``int``, *optional*
                The maximum length of the sentence,
                must not longer then the maximum sequence length for this model (i.e. ``tokenizer.model_max_length``).
            show_progress : ``int``, *optional*, defaults to True
                Show progress bar.
            pin_memory : ``bool``, *optional*, defaults to True
                Pin memory in order to accelerate the speed of data transfer to the GPU. This option is
                incompatible with multiprocessing.

        Returns
        -------
            ``List[List[str]]``
                A list of list of POS tags (``str``).
        """

        # Call model
        (
            logits,
            index_map,
        ) = super().__call__(input_text, use_delim=use_delim, **kwargs)

        # Get labels
        id2label = self.model.config.id2label

        # Post-process results
        output_text = []
        for sent_data in zip(input_text, index_map):
            output_sent = []
            for input_char, logits_index in zip(*sent_data):
                if logits_index is None or input_char.isspace():
                    label = "WHITESPACE"
                else:
                    label = id2label[np.argmax(logits[logits_index])]
                output_sent.append(label)
            output_text.append(output_sent)

        return output_text


################################################################################################################################


class CkipNerChunker(CkipTokenClassification):
    """The named-entity recognition driver.

    Parameters
    ----------
        model : ``str`` *optional*, defaults to "bert-base".
            The pretrained model name provided by CKIP Transformers.
        model_name : ``str`` *optional*, overwrites **model**
            The custom pretrained model name (e.g. ``'ckiplab/bert-base-chinese-ner'``).
        device : ``int``, *optional*, defaults to -1
            Device ordinal for CPU/GPU supports.
            Setting this to -1 will leverage CPU, a positive will run the model on the associated CUDA device id.
    """

    _model_names = {
        "albert-tiny": "ckiplab/albert-tiny-chinese-ner",
        "albert-base": "ckiplab/albert-base-chinese-ner",
        "bert-tiny": "ckiplab/bert-tiny-chinese-ner",
        "bert-base": "ckiplab/bert-base-chinese-ner",
    }

    def __init__(
        self,
        model: str = "bert-base",
        **kwargs,
    ):
        model_name = kwargs.pop("model_name", self._get_model_name(model))
        super().__init__(model_name=model_name, **kwargs)

    def __call__(
        self,
        input_text: List[str],
        *,
        use_delim: bool = False,
        **kwargs,
    ) -> List[List[NerToken]]:
        """Call the driver.

        Parameters
        ----------
            input_text : ``List[str]``
                The input sentences. Each sentence is a string.
            use_delim : ``bool``, *optional*, defaults to False
                Segment sentence (internally) using ``delim_set``.
            delim_set : `str`, *optional*, defaults to ``'，,。：:；;！!？?'``
                Used for sentence segmentation if ``use_delim=True``.
            batch_size : ``int``, *optional*, defaults to 256
                The size of mini-batch.
            max_length : ``int``, *optional*
                The maximum length of the sentence,
                must not longer then the maximum sequence length for this model (i.e. ``tokenizer.model_max_length``).
            show_progress : ``int``, *optional*, defaults to True
                Show progress bar.
            pin_memory : ``bool``, *optional*, defaults to True
                Pin memory in order to accelerate the speed of data transfer to the GPU. This option is
                incompatible with multiprocessing.

        Returns
        -------
            ``List[List[NerToken]]``
                A list of list of entities (:class:`~.util.NerToken`).
        """

        # Call model
        (
            logits,
            index_map,
        ) = super().__call__(input_text, use_delim=use_delim, **kwargs)

        # Get labels
        id2label = self.model.config.id2label

        # Post-process results
        output_text = []
        for sent_data in zip(input_text, index_map):
            output_sent = []
            entity_word = None
            entity_ner = None
            entity_idx0 = None
            for index_char, (
                input_char,
                logits_index,
            ) in enumerate(zip(*sent_data)):
                if logits_index is None:
                    label = "O"
                else:
                    label = id2label[np.argmax(logits[logits_index])]

                if label == "O":
                    entity_ner = None
                    continue

                bioes, ner = label.split("-")

                if bioes == "S":
                    output_sent.append(
                        NerToken(
                            word=input_char,
                            ner=ner,
                            idx=(
                                index_char,
                                index_char + len(input_char),
                            ),
                        )
                    )
                    entity_ner = None
                elif bioes == "B":
                    entity_word = input_char
                    entity_ner = ner
                    entity_idx0 = index_char
                elif bioes == "I":
                    if entity_ner == ner:
                        entity_word += input_char
                    else:
                        entity_ner = None
                elif bioes == "E":
                    if entity_ner == ner:
                        entity_word += input_char
                        output_sent.append(
                            NerToken(
                                word=entity_word,
                                ner=entity_ner,
                                idx=(
                                    entity_idx0,
                                    index_char + len(input_char),
                                ),
                            )
                        )
                    entity_ner = None

            output_text.append(output_sent)

        return output_text



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
The CKIP Transformers.
"""

__author_name__ = "Mu Yang"
__author_email__ = "emfomy@gmail.com"
__copyright__ = "2021 CKIP Lab"

__title__ = "CKIP Transformers"
__version__ = "0.3.0"
__description__ = "CKIP Transformers"
__license__ = "GPL-3.0"

__url__ = "https://ckip-transformers.readthedocs.io"
__download_url__ = __url__ + "/tarball/" + __version__



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = "Mu Yang <http://muyang.pro>"
__copyright__ = "2021 CKIP Lab"
__license__ = "GPL-3.0"

from setuptools import setup, find_namespace_packages
import ckip_transformers as about

################################################################################


def main():

    with open("README.rst", encoding="utf-8") as fin:
        readme = fin.read()

    setup(
        name="ckip-transformers",
        version=about.__version__,
        author=about.__author_name__,
        author_email=about.__author_email__,
        description=about.__description__,
        long_description=readme,
        long_description_content_type="text/x-rst",
        url=about.__url__,
        download_url=about.__download_url__,
        platforms=["linux_x86_64"],
        license=about.__license__,
        classifiers=[
            "Development Status :: 4 - Beta",
            "Environment :: Console",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3 :: Only",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Operating System :: POSIX :: Linux",
            "Natural Language :: Chinese (Traditional)",
        ],
        python_requires=">=3.6",
        packages=find_namespace_packages(
            include=[
                "ckip_transformers",
                "ckip_transformers.*",
            ]
        ),
        install_requires=[
            "torch>=1.5.0",
            "tqdm>=4.27",
            "transformers>=3.5.0",
        ],
        data_files=[],
    )


################################################################################

if __name__ == "__main__":
    main()
